{% extends "base.html" %}

{% block title %}{{ entry.title }}{% endblock %}

{% block content_title %}{{ entry.title }}{% endblock %}

{% block extra_header %}
  {% if session.logged_in %}
    <li><a href="{{ url_for('edit', slug=entry.slug) }}">Edit entry</a></li>
  {% endif %}
{% endblock %}

{% block content %}
  <p>Created {{ entry.timestamp.strftime('%m/%d/%Y at %G:%I%p') }}</p>
  {{ entry.html_content }}
{% endblock %}
