{% extends "base.html" %}

{% block title %}Log in{% endblock %}

{% block content_title %}Log in{% endblock %}

{% block content %}
  <form action="{{ url_for('login', next=next_url) }}" class="form-horizontal" method="post">
    <div class="form-group">
      <label for="password" class="col-sm-2 control-label">Password</label>
      <div class="col-sm-10">
        <input class="form-control" id="password" name="password" type="password">
      </div>
    </div>
    <div class="form-group">
      <div class="col-sm-offset-2 col-sm-10">
        <button class="btn btn-default" type="submit">Log in</button>
      </div>
    </div>
  </form>
{% endblock %}
